/*
 * FILE:        limited_heap.h
 *
 * AUTHORS:     Evgeniy Gabrilovich and Alex Gontmakher
 *              Technion - Israel Institute of Technology
 */


#ifndef __LIMITED_HEAP_H__
#define __LIMITED_HEAP_H__

#include <vector>
#include <utility>    // for 'pair'
#include <functional> // for 'greater'
#include <algorithm>  // for heap manipulation functions

using namespace std;


/* Limited size min-heap, which satisfies these conditions:
   1) The minimum element is at the root
   2) If an element to be added (via push_back) is smaller
      than the minimal element (residing at the root),
      the operation is aborted. 
      ==> It is this property that allows using this structure 
          for detecting the N largest elements of a sequence.
      ==> Using 'map' for this purpose is less space-efficient,
          since 'map' uses more storage per element than 'vector'
          (at least 2 pointers are needed per element to maintain
          a search tree).

   Template parameters:
   1) _Key - comparison is conducted according to this value
   2) _T   - payload (information accompanying the Key)
 */
template<class _Key, class _T>
class KMaxValues : public vector<pair<_Key, _T> > {
   typedef vector<pair<_Key, _T> > Base;

   // maximum size allowed
   int _n;

   /* make certain functions private, so that there is no need
      to override them to maintain the heap property */
   using Base::assign;
   using Base::erase;
   using Base::insert;
   using Base::pop_back;
   using Base::resize;
   using Base::swap;
   /* using op[] might inadvertently invalidate the heap property;
      better disable it and use iterators instead to access the elements */
   using Base::operator[];
   /* non-const iterators are similarly disabled */
   using Base::iterator;

   struct greater : public binary_function<value_type, value_type, bool> {
      bool operator()(const value_type& x, const value_type& y) const {
         return (x.first > y.first);
      }
    };
public:
   explicit KMaxValues(int maxSize = 1) 
      : _n(maxSize) { 
      /* Ensure there is room for at least 'maxSize' elements,
         without actually initializing them. That is, only
         'capacity()' is affected, while 'size()' is still 0. */
      reserve(maxSize);
      // there is no need to initialize the heap using 'make_heap',
      // as the structure is still empty
   }

   void push_back(const value_type& x) {
      if (size() < _n) {
         Base::push_back(x);
         push_heap(begin(), end(), greater());
      } else { // maximum size reached
         // check if the new element needs to be added at all,
         // by comparing it with the heap root
         if (x.first < begin()->first)
            return;
         
         // delete one element, then add the new one

         /* pop_heap moves the first element to the last position
            (i.e., vector[_n-1]), and then restores the heap
            property */
         pop_heap(begin(), end(), greater());
         /* The new value is just injected into the last position;
            using push_back would both unnecessarily grow the vector
            AND include anew the element that has been removed 
            by pop_heap. */
         (*this)[_n-1] = x;
         /* now include the newly added element into the heap */
         push_heap(begin(), end(), greater());
      }
   }
};


#endif /* __LIMITED_HEAP_H__ */
